/**
 * \file heartRate_driver.c
 * \version 1.00
 * \brief this file contains heart-rate functionality
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * Neither the name of NXP, nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * visit: http://www.mikroe.com and http://www.nxp.com
 *
 * get support at: http://www.mikroe.com/forum and https://community.nxp.com
 *
 * Project HEXIWEAR, 2015
 */

#include "heartrate_driver.h"
#include <math.h>
#include "arm_math.h"
#include "fsl_max.h"
#include "screens_resources.h"

guiImage_t screen_buttonStart = {.dynamicArea =
                                     {
                                         .xCrd = 50, .yCrd = 81,
                                     },

                                 .img = button_start_bmp};

guiImage_t screen_buttonStop = {.dynamicArea =
                                    {
                                        .xCrd = 50, .yCrd = 81,
                                    },

                                .img = button_stop_bmp};

guiLabel_t gui_heartRate_bpmVal_label = {.dynamicArea = {.xCrd = 33, .yCrd = 31, .width = 35, .height = 25},

                                         .textProperties = {.font = guiFont_18x20_Regular,
                                                            .fontColor = GUI_COLOR_WHITE,
                                                            .alignParam = OLED_TEXT_ALIGN_LEFT,
                                                            .background = NULL},

                                         .caption = NULL,
                                         .captionLength = 6};

guiLabel_t gui_heartRate_bpmText_label = {.dynamicArea = {.xCrd = 38, .yCrd = 55, .width = 28, .height = 14},

                                          .textProperties = {.font = guiFont_10x10_Regular,
                                                             .fontColor = GUI_COLOR_WHITE,
                                                             .alignParam = OLED_TEXT_ALIGN_LEFT,
                                                             .background = NULL},

                                          .caption = NULL,
                                          .captionLength = 6};

guiImage_t gui_heartRate_icon = {.dynamicArea = {.xCrd = 24, .yCrd = 23},

                                 .img = heartRate_icon_bmp};

/** private defines */
#define heartRate_IsPeak(array, idx) ((array[(idx)] > array[((idx) + 1)]) && (array[((idx))] > array[((idx)-1)]))
#define heartRate_UpdateFFT(size) (arm_rfft_fast_init_f32(&rfftInstance, size))
#define heartRate_InitFFT() (heartRate_UpdateFFT(MAXIM_INIT_SAMPLE_SIZE))

/** private variable definitions */
uint32_t volatile g_HeartRateNewVal = 0;
static uint32_t g_HeartRateoldVal = 0;

extern bool txOnGoing;
static uint16_t heartRate_currIdx = 0;

static uint16_t fft_currSize = MAXIM_INIT_SAMPLE_SIZE, fft_nextSize = MAXIM_INIT_SAMPLE_SIZE * 2;

static float32_t heartRate_datBuf[MAXIM_FINAL_SAMPLE_SIZE] = {0};

static arm_rfft_fast_instance_f32 rfftInstance;

static float32_t *fft_input = NULL, *fft_output = NULL;

static uint8_t oldHRVal = 0, heartRateVal = 0;
volatile bool g_heartUpdated = false;

/** private functions' declarations */
// static heartRate_status_t HeartRate_SendToKW40( uint32_t valToSend );
static void HeartRate_FindHR(float32_t *sigFFT, float32_t minFreq, float32_t maxFreq, uint8_t *heartRateVal);
static void HeartRate_UpdateHR(void);
static void HeartRate_ResetVars(void);

// extern uart_handle_t g_uartHandle;

/** Public API */
/**
 * register heart-rate sensor and FFT
 */
void HeartRate_Init()
{
    /* Gui object initialization */
    GuiDriver_LabelCreate(&gui_heartRate_bpmVal_label);
    snprintf((char *)gui_heartRate_bpmVal_label.caption, 4, "%03u", 000);

    GuiDriver_LabelCreate(&gui_heartRate_bpmText_label);
    GuiDriver_LabelSetCaption(&gui_heartRate_bpmText_label, (uint8_t *)"bpm");

    GuiDriver_ImageDraw(&gui_heartRate_icon);
    GuiDriver_ImageDraw(&screen_buttonStart);
}

/**
 * reallocate the buffers and reset the relevant variables
 */
status_t HeartRate_Resume()
{
    /** allocate input and output data buffers */

    fft_input = (float32_t *)malloc(MAXIM_FINAL_SAMPLE_SIZE * sizeof(*fft_input));
    if (NULL == fft_input)
    {
        return kStatus_Fail;
    }

    fft_output = (float32_t *)malloc((2 * MAXIM_FINAL_SAMPLE_SIZE) * sizeof(*fft_output));
    if (NULL == fft_output)
    {
        return kStatus_Fail;
    }

    HeartRate_ResetVars();

    return kStatus_Success;
}

/**
 * free used memory and unregister from HR sensor
 */
void HeartRate_Pause()
{
    /** deallocate data buffers */
    if (NULL != fft_input)
    {
        free(fft_input);
        fft_input = NULL;
    }

    if (NULL != fft_output)
    {
        free(fft_output);
        fft_output = NULL;
    }
}

void HeartRate_Process(void *sampleBuf, uint8_t numAvailSam, void *param)
{
    uint8_t *datBuf = (uint8_t *)sampleBuf;

    switch (numAvailSam)
    {
        case 0:
        {
            break;
        }

        /* overflow */
        case 0xFF:
        {
            HeartRate_ResetVars();
            break;
        }

        /* analyze data */
        default:
        {
            /* place available samples into buffer */
            for (uint8_t i = 0; i < numAvailSam; i++)
            {
                /** if it's time, update HR */
                if ((0 != heartRate_currIdx) && (0 == (heartRate_currIdx % MAXIM_NEW_SAMPLE_CHUNK)))
                {
                    HeartRate_UpdateHR();
                }

                /**
                 * if we reached the end of the buffer,
                 * shift data back to make room for new samples
                 */
                if (MAXIM_FINAL_SAMPLE_SIZE == heartRate_currIdx)
                {
                    memcpy((void *)&(heartRate_datBuf[0]),                                                // to
                           (const void *)&(heartRate_datBuf[MAXIM_NEW_SAMPLE_CHUNK]),                     // from
                           (MAXIM_FINAL_SAMPLE_SIZE - MAXIM_NEW_SAMPLE_CHUNK) * sizeof(*heartRate_datBuf) // how many
                           );
                    heartRate_currIdx = MAXIM_FINAL_SAMPLE_SIZE - MAXIM_NEW_SAMPLE_CHUNK;
                }

                /* process the new sample */
                uint32_t foo = (uint32_t)(((datBuf[i * MAXIM_BYTES_PER_ADC_VALUE + 0] << 16) |
                                           (datBuf[i * MAXIM_BYTES_PER_ADC_VALUE + 1] << 8) |
                                           datBuf[i * MAXIM_BYTES_PER_ADC_VALUE + 2]) &
                                          0x0003FFFF);
                if (foo)
                {
                    if (foo <= MAXIM_PROXY_THR)
                    {
                        /* Discard the samples */
                        HeartRate_ResetVars();
                    }
                    else
                    {
                        heartRate_datBuf[heartRate_currIdx] = (float32_t)foo;
                        heartRate_currIdx += 1;
                    }
                }
            }
        }
    }
}

/** private functions */

/**
 * reset relevant variables
 */
static void HeartRate_ResetVars()
{
    heartRate_currIdx = 0;
    fft_currSize = MAXIM_INIT_SAMPLE_SIZE;
    fft_nextSize = MAXIM_INIT_SAMPLE_SIZE * 2;
    heartRate_InitFFT();

    /** alert GUI */
    g_HeartRateNewVal = 0xFF;
}

/**
 * calculate and update current heart rate
 */
static void HeartRate_UpdateHR()
{
    /* update FFT size if applicable */
    if ((fft_nextSize <= MAXIM_FINAL_SAMPLE_SIZE) && (heartRate_currIdx >= fft_nextSize))
    {
        fft_currSize = fft_nextSize;
        fft_nextSize *= 2;
        heartRate_UpdateFFT(fft_currSize);
    }

    /* copy temporal data into a temporary buffer, as FFT will change the input array */
    memcpy((void *)fft_input, (const void *)&(heartRate_datBuf[heartRate_currIdx - fft_currSize]),
           fft_currSize * sizeof(*fft_input));

    /* calculate FFT */
    arm_rfft_fast_f32(&rfftInstance, fft_input, fft_output, 0);

    /* calculate spectral magnitude values */
    arm_cmplx_mag_f32(fft_output, fft_input, fft_currSize);

    oldHRVal = heartRateVal;

    /* use these magnitude values to find the dominant frequency peak */
    HeartRate_FindHR(fft_input, MAXIM_HR_FREQ_MIN, MAXIM_HR_FREQ_MAX, &heartRateVal);

    if ((0xFF != heartRateVal) && (oldHRVal != heartRateVal))
    {
        /* send to GUI app */
        g_HeartRateNewVal = heartRateVal;
        g_heartUpdated = true;
    }
}

/**
 * calculate heart-rate by finding the biggest peak
 * in the spectrum
 *
 * @param sigFFT       data FFT
 * @param minFreq      min frequency border
 * @param maxFreq      max frequency border
 * @param heartRateVal calculated value returned
 */
static void HeartRate_FindHR(float32_t *sigFFT, float32_t minFreq, float32_t maxFreq, uint8_t *heartRateVal)
{
    uint16_t minIdx = (uint16_t)floor(minFreq * fft_currSize / HR_SAMPLE_RATE),
             maxIdx = (uint16_t)ceil(maxFreq * fft_currSize / HR_SAMPLE_RATE);

    uint16_t currPeak;

    uint32_t peaks[MAXIM_FIFO_DEPTH / 2] = {0};

    /**
     * find all the peaks in the [ 45 - 180 ] bpm range,
     * then find the biggest one among them
     */

    uint32_t peakIdx = 0;

    for (uint16_t i = minIdx; i <= maxIdx; i++)
    {
        if (0 != heartRate_IsPeak(sigFFT, i))
        {
            peaks[peakIdx] = i;
            peakIdx += 1;
        }
    }

    /* find the biggest among all the found peaks */
    if (peakIdx > 0)
    {
        currPeak = peaks[0];
        for (uint16_t i = 0; i < peakIdx; i++)
        {
            if (sigFFT[currPeak] < sigFFT[peaks[i]])
            {
                currPeak = peaks[i];
            }
        }

        /* convert to BPM */
        *heartRateVal = (uint8_t)(60 * (((float32_t)currPeak * HR_SAMPLE_RATE) / fft_currSize));
    }
    else
    {
        *heartRateVal = 0xFF;
    }
}

void HeartRate_UpateGUI(void)
{
    GuiDriver_LabelDraw(&gui_heartRate_bpmVal_label);
    GuiDriver_LabelDraw(&gui_heartRate_bpmText_label);
    if (g_HeartRateNewVal != g_HeartRateoldVal)
    {
        uint16_t valToWrite = 0;

        if (g_HeartRateNewVal == 0xFF)
        {
            valToWrite = 0;
        }
        else
        {
            valToWrite = g_HeartRateNewVal;
        }

        g_HeartRateoldVal = g_HeartRateNewVal;

        sprintf((char *)gui_heartRate_bpmVal_label.caption, "%03u", valToWrite);
        GuiDriver_LabelDraw(&gui_heartRate_bpmVal_label);
    }
}

void HeartRate_ResumeGUI(void)
{
    GuiDriver_CleanMainArea();
    GuiDriver_LabelDraw(&gui_heartRate_bpmVal_label);
    GuiDriver_LabelDraw(&gui_heartRate_bpmText_label);
    GuiDriver_ImageDraw(&screen_buttonStop);
    HeartRate_Resume();
}

void HeartRate_PauseGUI(void)
{
    HeartRate_Pause();

    GuiDriver_CleanMainArea();
    GuiDriver_ImageDraw(&gui_heartRate_icon);
    GuiDriver_ImageDraw(&screen_buttonStart);
}
